package com.ibm.jps.uxtoweb.frontend.custom;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.uml2.uml.AggregationKind;
import org.eclipse.uml2.uml.Association;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Comment;
import org.eclipse.uml2.uml.Component;
import org.eclipse.uml2.uml.Dependency;
import org.eclipse.uml2.uml.Element;
import org.eclipse.uml2.uml.Model;
import org.eclipse.uml2.uml.NamedElement;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.Package;
import org.eclipse.uml2.uml.Property;
import org.eclipse.uml2.uml.Type;

public class ModelUtil {
	private static Class foundClass = null;
	private static Association foundAssoc = null;
	private static Property foundRole = null;
	
	public static String getNameWithoutModel(String name) {
		int pos = name.indexOf("::"); //$NON-NLS-1$
		if (pos > 0)
			return name.substring(name.indexOf("::") + 2); //$NON-NLS-1$
		else
			return name;
	}

	public static String convertToFullJavaName(String input) {
		return getNameWithoutModel(input).replaceAll("::", "."); //$NON-NLS-1$ //$NON-NLS-2$
	}

	public static String convertToSimpleJavaName(String input) {
		return getSimpleName(convertToFullJavaName(input));
	}

	public static String convertToJavaPath(String input) {
		return getPath(convertToFullJavaName(input));
	}

    /**
     * Given an input string of the type "com.company.project.AClass", getPath will return
     * "com.company.project" (or an empty string if the input string is "AClass").
     * 
     * @param name
     * @return
     */
    public static String getPath(String name) {
        int pos = name.lastIndexOf("."); //$NON-NLS-1$
        if (pos > 0)
            return name.substring(0, pos);
        else
            return ""; //$NON-NLS-1$
    }

    public static String getPath(String name, char fileSeparator) {
        int pos = name.lastIndexOf(fileSeparator);
        if (pos > 0)
            return name.substring(0, pos);
        else
            return ""; //$NON-NLS-1$
    }

    /**
     * Given an input string of the type "com.company.project.AClass", getSimpleName will return
     * "AClass" (or the string itself if the input string is "AClass").
     * 
     * @param name
     * @return
     */
    public static String getSimpleName(String name) {
        int pos = name.lastIndexOf("."); //$NON-NLS-1$
        if (pos > 0)
            return name.substring(pos + 1);
        else
            return name;
    }

    public static String getSimpleName(String name, char fileSeparator) {
        int pos = name.lastIndexOf(fileSeparator);
        if (pos > 0)
            return name.substring(pos + 1);
        else
            return name;
    }

	public static String getDocumentation(Element from) {
		for (Iterator i = from.getOwnedComments().iterator(); i.hasNext();) {
			Comment c = (Comment) i.next();
			if (c.getAppliedStereotype(Messages.documentationStereotype) != null) { 
				return c.getBody();
			}
		}
		return ""; //$NON-NLS-1$
	}
	
	public static IProject getProject(NamedElement elt) {
		Model model = elt.getModel();
		Resource r = model.eResource();
		if (r != null) {
			URI uri = r.getURI();
			IWorkspaceRoot wr = ResourcesPlugin.getWorkspace().getRoot();
			String pName = uri.path().substring("/resource".length() + 1); //$NON-NLS-1$
			int pos = pName.indexOf("/"); //$NON-NLS-1$
			if (pos > 0) {
				pName = pName.substring(0, pos);
				return wr.getProject(pName);
			}
		}
		return null;
	}

    public static IResource getModelResource(NamedElement elt) {
        Model model = elt.getModel();
        Resource r = model.eResource();
//        if (r != null) {
//            URI uri = r.getURI();
//            IWorkspaceRoot wr = ResourcesPlugin.getWorkspace().getRoot();
//            String pName = uri.path().substring("/resource".length() + 1); //$NON-NLS-1$
//            String mName = null;
//            int pos = pName.indexOf("/"); //$NON-NLS-1$
//            if (pos > 0) {
//                mName = pName.substring(pos + 1);
//                pName = pName.substring(0, pos);
//                IProject prj = wr.getProject(pName);
//                if (prj != null) {
//                    return prj.findMember(mName);
//                }
//            }
//        }

        if (r != null) {
			IWorkspaceRoot wr = ResourcesPlugin.getWorkspace().getRoot();
			String uri = r.getURI().path().replaceAll("%20", " ");
			if (uri.startsWith("/resource/")) {
				String pName = uri.substring("/resource".length() + 1);
				String mName = null;
				int pos = pName.indexOf("/");
				if (pos > 0) {
					mName = pName.substring(pos + 1);
					pName = pName.substring(0, pos);
					IProject prj = wr.getProject(pName);
					if (prj != null) {
						return prj.findMember(mName);
					}
				}
			}
		}
        return null;
    }
    
    public static IMarker addMarker(NamedElement elt, String description, int severity) {
        System.err.println(description);
		IResource onRsrc = getModelResource(elt);
        if (onRsrc == null) return null;
        try {
            IMarker marker = onRsrc.createMarker(IMarker.PROBLEM);
            if (marker.exists()) {
                marker.setAttribute(IMarker.MESSAGE, description);
                marker.setAttribute(IMarker.SEVERITY, severity);
                marker.setAttribute(IMarker.TRANSIENT, true);
                return marker;
            }
        } catch (CoreException e) {
            System.err.println("Marker does not exist."); //$NON-NLS-1$
        }
        return null;
    }
    
	public static Class findMatchingBean(Class fromClass, String attribName) {
		for (Object curDep_o:fromClass.getClientDependencies()) {
			Dependency curDep = (Dependency) curDep_o;
			for (Object curElt_o:curDep.getSuppliers()) {
				NamedElement curElt = (NamedElement) curElt_o;
				if (curElt instanceof Class) {
					Class supplier = (Class) curElt;
					if (supplier.getAppliedStereotype(Messages.beanStereotype) != null) {
						for (Object curProp_o:supplier.getAttributes()) {
							Property curProp = (Property) curProp_o;
							if (curProp.getName().equals(attribName)) {
								return supplier;
							}
						}
					}
				}
			}
		}
		return null;
	}
	
	private static String findMatchingBeanNameFromDependencies(Operation inOp) {
		Class cl = (Class) inOp.getNamespace();
		// First find the list of beans associated with this class 
		List<SupplierClass> beans = new ArrayList<SupplierClass>();
		for (Object curDep_o:cl.getClientDependencies()) {
			Dependency curDep = (Dependency) curDep_o;
			for (Object curElt_o:curDep.getSuppliers()) {
				NamedElement curElt = (NamedElement) curElt_o;
				if (curElt instanceof Class) {
					Class supplier = (Class) curElt;
					if (supplier.getAppliedStereotype(Messages.beanStereotype) != null) {
						SupplierClass sup = new SupplierClass();
						sup.setDep(curDep);
						sup.setSupplier(supplier);
						beans.add(sup);
					}
				}
			}
		}
		if (beans.size() == 1) {
			// Since there is only one dependency, it should either not be named, or include the operation
			// name in its own name
			Dependency dep = beans.get(0).getDep();
			if (dep.getName() == null || dep.equals("") || relNameMatches(dep.getName(), inOp.getName())) {
				return beans.get(0).getSupplier().getName();
			}
		} else if (beans.size() > 1) {
			// Since there are several possible choices, the match should be by name
			for (SupplierClass sup:beans) {
				Dependency dep = sup.getDep();
				if (dep.getName() != null &&  relNameMatches(dep.getName(), inOp.getName())) {
					return sup.getSupplier().getName();
				}
			}
		}
		return "";
	}
	
	/**
	 * Finds the association named 'assocName' of type specified by 'aKind' for
	 * class 'fromClass'
	 * 
	 * @param fromClass
	 *            the origin of the association
	 * @param targetClassStereotype
	 *            the stereotype of the target class (if specified)
	 * @param aKind
	 *            the type of the association to find (aggregation, composition
	 *            or straight association) (if specified)
	 * @param assocName
	 *            the name of the association to find (if specified)
	 */
	private static void findIt(Class fromClass, String targetClassStereotype, AggregationKind aKind, String assocName) {
		foundAssoc = null;
		foundClass = null;
		// Search the package of the specified class
		Package pkg = fromClass.getPackage();
		if (pkg != null) {
			searchPackage(pkg, fromClass, targetClassStereotype, aKind, assocName);
		}
		// Search the whole model if not found
		if (foundClass == null) {
			Model model = fromClass.getModel();
			searchPackage(model, fromClass, targetClassStereotype, aKind, assocName);
		}
	}

	/**
	 * @param pkg
	 * @param fromClass
	 * @param targetClassStereotype
	 * @param kind
	 * @param inRefName
	 */
	private static void searchPackage(Package pkg, Class fromClass, String targetClassStereotype, AggregationKind aKind,
			String inRefName) {
		for (Object o:pkg.getOwnedElements()) {
			Element ownedElt = (Element) o;
			if (ownedElt instanceof Association) {
				Association assoc = (Association) ownedElt;
				EList members = assoc.getMemberEnds();
				if (members.size() >= 2) {
					Property prop1 = (Property) members.get(0);
					Property prop2 = (Property) members.get(1);
					Type type1 = prop1.getType();
					Type type2 = prop2.getType();
					if (type1 instanceof Class && type2 instanceof Class) {
						Class cl1 = (Class) type1;
						Class cl2 = (Class) type2;
						Class targetClass = null;
						Property prop = null;
						if (cl1.equals(fromClass)) {
							targetClass = cl2;
							prop = prop2;
						} else if (cl2.equals(fromClass)) {
							targetClass = cl1;
							prop = prop1;
						}
						// Process further only if one of the ends matches fromClass and if the targetClass 
						// is navigable (non-navigable ends are owned by the association itself instead of 
						// being owned by the class)
						if (prop != null && !assoc.getOwnedEnds().contains(prop)) {
							if (aKind == null || prop.getAggregation().equals(aKind)) {
								if (inRefName == null || inRefName.equals("") || relNameMatches(assoc.getName(), inRefName)) { //$NON-NLS-1$
									foundAssoc = assoc;
									if (targetClassStereotype == null || targetClassStereotype.equals("") ||  //$NON-NLS-1$
										targetClass.getAppliedStereotype(targetClassStereotype) != null) {
										foundClass = targetClass;
										foundRole = prop;
										return;
									}
								}
							}
						}
					}
				}
			}
		}
		// Search the package nested packages
		for (Object o:pkg.getNestedPackages()) {
			searchPackage((Package) o, fromClass, targetClassStereotype, aKind, inRefName);
			if (foundClass != null)return;
		}
	}
	
	private static boolean relNameMatches(String inRelName, String inRefName) {
		if (inRelName.equals(inRefName)) {
			return true;
		} else {
			// E.g. assocName = 'addCd, updateCd' and refName = 'addCd' 
			String [] tokens = inRelName.split(",");
			for (String token:tokens) {
				if (token.trim().equals(inRefName)) return true;
			}
		}
		return false;
	}

	public static AssociatedClass findCompositionClass(Class fromClass, String targetClassStereotype) {
		findIt(fromClass, targetClassStereotype, AggregationKind.get(AggregationKind.COMPOSITE), null);
		return new AssociatedClass(foundClass, foundAssoc, foundRole);
	}
	
    public static String convertToTitle(String inputString) {
        String outputString = "";
        int cc = inputString.length();
        for (int i = 0; i < cc; i++) {
            char c = inputString.charAt(i);
            boolean eos = true;
            char cn = 0;
            if (i < cc - 1) {
                eos = false;
                cn = inputString.charAt(i + 1);
            }
            if (i == 0)
                outputString = outputString + Character.toUpperCase(c);
            else {
                if (Character.isUpperCase(c) && !eos && !Character.isUpperCase(cn) && cn != ' '
                        && !outputString.endsWith(" "))
                    outputString = outputString + " " + c;
                else
                    outputString = outputString + c;
            }
        }
        return outputString;
    }

	public static Class findAssocClass(Class fromClass, String targetClassStereotype, AggregationKind aKind) {
		findIt(fromClass, targetClassStereotype, aKind, null);
		return foundClass;
	}
	
	public static Class findAssocClass(Class fromClass, String targetClassStereotype, AggregationKind aKind, String assocName) {
		findIt(fromClass, targetClassStereotype, aKind, assocName);
		return foundClass;
	}
	
	public static Class findNavigableCompositionEnd(Property prop){
		Association assoc = prop.getAssociation();
		if (assoc != null) {
			for (Object o:assoc.getMemberEnds()) {
				Property pr = (Property) o;
				Type tp = (Type) pr.getType();
				// Look for the same role name
				if (pr.getName() != null && pr.getName().equals(prop.getName())) {
					// Must be navigable
					if (!assoc.getOwnedEnds().contains(pr)) {
						// Must be a composite
						if (pr.getAggregation().equals(AggregationKind.COMPOSITE_LITERAL)) {
							// Must be a Class
							if (tp != null && tp instanceof Class) return (Class) tp;
						}
					}
				}
			}
		}
		return null;
	}
	
	public static Class findNavigableAssociationEnd(Property prop){
		Association assoc = prop.getAssociation();
		if (assoc != null) {
			for (Object o:assoc.getMemberEnds()) {
				Property pr = (Property) o;
				Type tp = (Type) pr.getType();
				// Look for the same role name
				if (pr.getName() != null && pr.getName().equals(prop.getName())) {
					// Must be navigable
					if (!assoc.getOwnedEnds().contains(pr)) {
						// Must be a composite
						if (pr.getAggregation().equals(AggregationKind.NONE_LITERAL)) {
							// Must be a Class
							if (tp != null && tp instanceof Class) return (Class) tp;
						}
					}
				}
			}
		}
		return null;
	}
	
    /**
     * @param name
     * @return
     */
    public static String firstCharToUpperCase(String name) {
        return name.substring(0, 1).toUpperCase() + name.substring(1);
    }

    /**
     * @param name
     * @return
     */
    public static String firstCharToLowerCase(String name) {
        return name.substring(0, 1).toLowerCase() + name.substring(1);
    }

	/**
	 * Extract classes recursively starting from the input package. A stereotype may be specified.
	 * 
	 * @param inPkg
	 * @param inStereo
	 * @param inList
	 */
	public static void extractClasses(Package inPkg, String inStereo, List<Class> inList) {
		for (Object curElt_o : inPkg.getOwnedElements()) {
			Element curElt = (Element) curElt_o;
			if (curElt instanceof Class) {
				Class cl = (Class) curElt;
				if (inStereo.equals("") || cl.getAppliedStereotype(inStereo) != null) {
					inList.add(cl);
				}
			}
		}
		for (Object curPkg_o:inPkg.getNestedPackages()) {
			extractClasses((Package) curPkg_o, inStereo, inList);
		}
	}

	/**
	 * Extract components recursively starting from the input package. A stereotype may be specified.
	 * 
	 * @param inPkg
	 * @param inStereo
	 * @param inList
	 */
	public static void extractComponents(Package inPkg, String inStereo, List<Component> inList) {
		for (Object curElt_o : inPkg.getOwnedElements()) {
			Element curElt = (Element) curElt_o;
			if (curElt instanceof Component) {
				Component comp = (Component) curElt;
				if (inStereo.equals("") || comp.getAppliedStereotype(inStereo) != null) {
					inList.add(comp);
				}
			}
		}
		for (Object curPkg_o:inPkg.getNestedPackages()) {
			extractComponents((Package) curPkg_o, inStereo, inList);
		}
	}
	
	public static void log(String from, EObject arg0) {
		System.out.println(from + ": " + objectheader(arg0));
	}
	
	public static void log(String from, Object arg0) {
		if (arg0 instanceof EObject) {
			System.out.println(from + ": " + objectheader((EObject) arg0));
		} else {
			System.out.println(from + ": " + arg0.getClass().getSimpleName());
		}
	}
	
	public static void log(String from, EObject arg0, EObject arg1) {
		System.out.println(from + ": " + objectheader(arg0) + "; " + objectheader(arg1));
	}
	
	private static String objectheader (EObject obj) {
		String s = obj.eClass().getName() + " ";
		if (obj instanceof NamedElement) {
			s += ((NamedElement) obj).getName();
		} else {
			s += " (not a named element)";
		}
		return s;
	}
	
	public static List<AssociatedClass> findAssociations(Class fromClass, String refName) {
		List<AssociatedClass> assocs = new ArrayList<AssociatedClass>();
		Package pkg = fromClass.getModel();
		if (pkg != null) {
			addAssociations(assocs, pkg, fromClass, refName);
		}
		return assocs;
	}

	private static void addAssociations(List<AssociatedClass> toList, Package inPkg, Class fromClass, String refName) {
		for (Object o:inPkg.getOwnedElements()) {
			Element ownedElt = (Element) o;
			if (ownedElt instanceof Association) {
				Association assoc = (Association) ownedElt;
				EList members = assoc.getMemberEnds();
				if (members.size() >= 2) {
					Property prop1 = (Property) members.get(0);
					Property prop2 = (Property) members.get(1);
					Type type1 = prop1.getType();
					Type type2 = prop2.getType();
					if (type1 instanceof Class && type2 instanceof Class) {
						Class cl1 = (Class) type1;
						Class cl2 = (Class) type2;
						if (cl1.equals(fromClass) || cl2.equals(fromClass)) {
							boolean addIt = false;
							String assocName = assoc.getName();
							if (assocName != null) {
								if (refName.equals("")) {
									if (assocName.startsWith("[")) addIt = true;
								} else {
									if (assocName.equals(refName) || assocName.startsWith(refName + " [")) {
										addIt = true;
									} else {
										String s = assocName;
										int pos = assocName.indexOf("[");
										if (pos > 0) {
											s = assocName.substring(0, pos);
										}
										s = s.trim();
										if (s.startsWith(refName + ",") || s.endsWith(", " + refName) || s.indexOf(", " + refName + ",") > 0) {
											addIt = true;
										}
									}
								}
							}
							if (addIt) {
								if (cl1.equals(fromClass)) {
									toList.add(new AssociatedClass(cl2, assoc, prop2));
								} else {
									toList.add(new AssociatedClass(cl1, assoc, prop1));
								}
							}
						}
					}
				}
			}
		}
		// Search the nested packages
		for (Object o:inPkg.getNestedPackages()) {
			addAssociations(toList, (Package) o, fromClass, refName);
		}
	}

	private static int indexNextUpperCase(String inStr, int inStart) {
        int cc = inStr.length();
        for (int i = inStart; i < cc; i++) {
            char c = inStr.charAt(i);
            if (Character.isUpperCase(c)) {
            	return i;
            }
        }
        return -1;
	}
	
	private static boolean matchingClass(String inName, List<Class> inList) {
		for (Class cl:inList) {
			if (cl.getName().equals(inName)) return true;
		}
		return false;
	}
	
	/**
	 * findMatchingBeanName extracts from the input operation the name of a valid <<bean>> class. The format for
	 * the operation name is: <keyword><beanName><optionalText> in CamelCase. A <keyword> is a single word starting
	 * in lower case. The operation can be a <<useraction>> or a 'list' <<link>>.
	 * 
	 * @param inOp
	 * @return
	 */
	public static String findMatchingBeanName(Operation inOp) {
		// Retrieve all valid <<bean>> classes
		List<Class> beans = new ArrayList<Class>();
		extractClasses(inOp.getModel(), Messages.beanStereotype, beans);
		// Identify the keyword
		String opName = inOp.getName();
		int beanNameStart = indexNextUpperCase(opName, 0);
		if (beanNameStart != -1) {
			String beanName = opName.substring(beanNameStart);
			// In most cases (User from findUser, Cd from addCd, etc.) the bean name is the extracted value
			if (matchingClass(beanName, beans)) return beanName;
			// No match (e.g. UserByName from findUserByName): go through the list of beans
			String foundName = "";
			for (Class cl:beans) {
				if (beanName.startsWith(cl.getName())) {
					// E.g. UserByName starts with name of class User
					// To be valid, the next character (here B) must be in uppercase
					// This algorithm looks for the longest possible bean name if several beans are
					// valid. E.g. operation findPayCheckByDate and beans Pay, PayCheck. If Pay is 
					// picked first, the search would return the wrong bean name.
					if (Character.isUpperCase(beanName.charAt(cl.getName().length()))) {
						if (cl.getName().length() > foundName.length()) foundName = cl.getName();
					}
				}
			}
			if (foundName.length() > 0) return foundName;
			// Still no match: look for a dependency 
			return findMatchingBeanNameFromDependencies(inOp);
		} else {
			return findMatchingBeanNameFromDependencies(inOp);
		}
	}
}
