package com.ibm.jps.uxtoweb.frontend.custom;

import inputSchema.FindOp;
import inputSchema.FindParam;
import inputSchema.InputSchemaFactory;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.ConnectableElement;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.Property;
import org.eclipse.uml2.uml.Stereotype;

import com.ibm.xtools.transform.authoring.RuleExtension;

public class Operation2FindOpRule implements RuleExtension {

	@SuppressWarnings("unchecked")
	public void execute(EObject arg0, EObject arg1) {
		ModelUtil.log("Operation2FindOpRule", arg0);
		Operation inOp = (Operation) arg0;
		FindOp inFindOp= (FindOp) arg1;
		Class inCl = (Class) inOp.getNamespace();
		// inOp is guaranteed to be a <<useraction>> operation named 'find<beanName>'. The only thing we need to do is
		// to retrieve the FindParam by identifying the <<input>> class associated by composition with the parent class
		// of the <<useraction> operation. Once found, the <<...>> fields of the <<input>> class become the FindParam. 
		for (Object curElt_o:inCl.getRoles()) {
			ConnectableElement curElt = (ConnectableElement) curElt_o;
			// A <<screen>> may point to an <<input>> form using a composition. Note that we access the <<imput>> 
			// form though the role (property) of the navigable composition. 
			if (curElt instanceof Property) {
				Class cl = ModelUtil.findNavigableCompositionEnd((Property) curElt);
				if (cl != null && cl.getAppliedStereotype(Messages.inputStereotype) != null) {
					// We have identified an <<input>> class: cycle through its attributes. Candidate attributes have
					// specific stereotypes (like <<inputtext>>).
					for (Object curProp_o:cl.getAllAttributes()) {
						Property curProp = (Property) curProp_o;
						for (Object curSt_o:curProp.getAppliedStereotypes()) {
							Stereotype curSt = (Stereotype) curSt_o;
							String curStName = curSt.getQualifiedName();
							if (curStName.equals(Messages.inputtextStereotype) || 
								curStName.equals(Messages.listboxStereotype) || 
								curStName.equals(Messages.radiobuttonStereotype)) {
								// We have a candidate attribute. Only retain attributes that match our operation.
								Class bean = ModelUtil.findMatchingBean(cl, curProp.getName());
								if (bean != null) {
									String refName = "find" + bean.getName();
									if (inOp.getName().startsWith(refName)) {
										FindParam findParam = InputSchemaFactory.eINSTANCE.createFindParam();
										findParam.setName(curProp.getName());
										inFindOp.getFindParam().add(findParam);
										if (curProp.getType() == null || !(curProp.getType().getName().equals(Messages.stringLabel))) {
//							                ModelUtil.addMarker(curProp, "In the current version, elements to search must be of type String. " + 
//							                		"Find/Searches involving the attribute " + curProp.getName() + 
//							                		" (in " + curProp.getNamespace().getName() + ") will not work.",
//													IMarker.SEVERITY_WARNING);
										}
										break;
									}
								}
							}
						}
					}
				}
			}		
		}
	}

}
