package com.ibm.jps.uxtoweb.frontend.custom;

import inputSchema.InputSchemaFactory;
import inputSchema.Logo;
import inputSchema.MainEntryPoint;
import inputSchema.Webproject;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Dependency;
import org.eclipse.uml2.uml.Model;
import org.eclipse.uml2.uml.NamedElement;
import org.eclipse.uml2.uml.Stereotype;

import com.ibm.xtools.transform.authoring.RuleExtension;

public class Model2WebprojectRule implements RuleExtension {

	@SuppressWarnings("unchecked") //$NON-NLS-1$
	public void execute(EObject arg0, EObject arg1) {
		ModelUtil.log("Model2WebprojectRule", arg0); //$NON-NLS-1$
		Model model = (Model) arg0;
		Webproject proj = (Webproject) arg1;
		
		// The project name is initialized with the model name but will be overwritten if there are any <<settings>>
		// classes
		proj.setName(model.getName());
		
		// appName and logo are stored in <<settings>> classes

		proj.setAppName(model.getName()); // default
		List<Class> projClasses = new ArrayList<Class>();
		ModelUtil.extractClasses(model, Messages.settingsStereotype, projClasses);
		for (Class cl:projClasses) {
			Stereotype st = cl.getAppliedStereotype(Messages.settingsStereotype);
			if (st != null) {
				// Project name
				proj.setName(cl.getName());
				// Webproject appName
				String s = (String) cl.getValue(st, Messages.appnameProperty); //$NON-NLS-1$
				if (s != null && !s.equals("")) proj.setAppName(s); //$NON-NLS-1$
				// Webproject logo
	            String logoFName = (String) cl.getValue(st, Messages.logoFile);
	            if (logoFName != null && !logoFName.equals("")) { //$NON-NLS-1$
	                IProject prj = ModelUtil.getProject(model);
	                IFile f = prj.getFile(logoFName);
	                if (!f.exists()) {
	                    ModelUtil.addMarker(model, "Invalid logo file name " + logoFName //$NON-NLS-1$
	                             + ". The logo file should be located in the same project as the model itself. Examples: '" //$NON-NLS-1$
	                             + "logofile.gif' or 'subdir\\logofile.jpg'.", IMarker.SEVERITY_ERROR); //$NON-NLS-1$
	                } else {
	                    Logo logo = InputSchemaFactory.eINSTANCE.createLogo();
	                    proj.getLogo().add(logo);
	                    logo.setFileName(ModelUtil.getSimpleName(f.getFullPath().toOSString(), File.separatorChar));
	                    logo.setFilePath(ModelUtil.getPath(f.getFullPath().toOSString(), File.separatorChar).substring(1));
	                    Integer temp = (Integer) cl.getValue(st, Messages.logoWidth);
	                    if (temp != null) logo.setWidthInPixels(temp.toString());
	                    temp = (Integer) cl.getValue(st, Messages.logoHeight);
	                    if (temp != null) logo.setHeightInPixels(temp.toString());
	                }
	            }
	            // Main Entry Point: look for dependencies to <<screen>> classes. Only one entry point is allowed.
	    		String mainEntryPointName = "";
	    		for (Object curDep_o:cl.getClientDependencies()) {
	    			Dependency curDep = (Dependency) curDep_o;
	    			for (Object curElt_o:curDep.getSuppliers()) {
	    				NamedElement curElt = (NamedElement) curElt_o;
	    				if (curElt instanceof Class) {
	    					Class supplier = (Class) curElt;
	    					if (supplier.getAppliedStereotype(Messages.screenStereotype) != null) {
								if (mainEntryPointName.equals("")) {
									mainEntryPointName = supplier.getName();
									MainEntryPoint entryPoint = InputSchemaFactory.eINSTANCE.createMainEntryPoint();
									entryPoint.setName(ModelUtil.firstCharToLowerCase(mainEntryPointName));
									proj.getMainEntryPoint().add(entryPoint);
								} else {
				                    ModelUtil.addMarker(model, "A project can have only one entry point. " //$NON-NLS-1$
				                             + "The entry point is already defined as '" + mainEntryPointName + "'. The dependency " //$NON-NLS-1$
				                             + "from <<settings>> class " + cl.getName() + " to <<screen>> class " + supplier.getName() 
				                             + " has been ignored and should be removed from the model.", IMarker.SEVERITY_ERROR); //$NON-NLS-1$
								}
							}
	    				}
	    			}
	    		}
			}
		}
	}
}
